open StdLabels

module Helper = struct
  type type_of =
    | Integer  (** A numeric value *)
    | Bool  (** A boolean, not a real type  *)
    | String  (** String value *)
    | NumericString
        [@printer fun fmt _ -> Format.pp_print_string fmt "Integer"]
        (** String containing a numeric value *)
  [@@deriving show { with_path = false }]

  type t = Variable of type_of | Raw of type_of
  type argument_repr = { pos : S.pos; t : t }

  module DynType = struct
    type nonrec t = t -> t
    (** Dynamic type is a type unknown during the code.

      For example, the equality operator accept either Integer or String, but
      we expect that both sides of the equality uses the same type.*)

    (** Build a new dynamic type *)
    let t : unit -> t =
     fun () ->
      let stored = ref None in
      fun t ->
        match !stored with
        | None ->
            stored := Some t;
            t
        | Some t -> t
  end

  (** Declare an argument for a function. 

 - Either we already know the type and we just have to compare.
 - Either the type shall constrained by another one 
 - Or we have a variable number of arguments. *)
  type argument =
    | Fixed of type_of
    | Dynamic of DynType.t
    | Variable of argument

  let compare :
      ?strict:bool ->
      ?level:Report.level ->
      type_of ->
      argument_repr ->
      Report.t list ->
      Report.t list =
   fun ?(strict = false) ?(level = Report.Warn) expected actual report ->
    let equal =
      match (expected, actual.t) with
      (* Strict equality for this ones, always true *)
      | String, Variable String
      | String, Raw String
      | String, Raw NumericString
      | String, Variable NumericString
      | Integer, Variable Integer
      | Integer, Raw Integer
      | NumericString, Raw NumericString
      | NumericString, Variable NumericString
      | Bool, Raw Bool
      | Bool, Variable Bool
      (* Also include the conversion between bool and integer *)
      | Integer, Raw Bool
      | Integer, Variable Bool
      (* The type NumericString can be used as a generic type in input *)
      | _, Variable NumericString
      | NumericString, Raw String
      | NumericString, Variable String
      | NumericString, Raw Integer
      | NumericString, Variable Integer
      (* A numeric type can be used at any place *)
      | String, Raw Integer ->
          true
      | Bool, Variable Integer when not strict -> true
      | Bool, Raw Integer when not strict -> true
      | String, Variable Integer when not strict -> true
      | String, Raw Bool when not strict -> true
      | String, Variable Bool when not strict -> true
      | Integer, Variable String when not strict -> true
      (* Explicit rejected cases  *)
      | Integer, Raw NumericString when not strict -> true
      | Integer, Raw String -> false
      | _, _ -> false
    in
    if equal then report
    else
      let result_type = match actual.t with Variable v -> v | Raw r -> r in
      let message =
        Format.asprintf "The type %a is expected but got %a" pp_type_of expected
          pp_type_of result_type
      in
      Report.message level actual.pos message :: report

  let rec compare_parameter :
      ?strict:bool ->
      ?level:Report.level ->
      argument ->
      argument_repr ->
      Report.t list ->
      Report.t list =
   fun ?(strict = false) ?(level = Report.Warn) expected param report ->
    match expected with
    | Fixed t -> compare ~level t param report
    | Dynamic d ->
        let type_ = match d param.t with Raw r -> r | Variable v -> v in
        compare ~strict ~level type_ param report
    | Variable c -> compare_parameter ~level c param report

  (** Compare the arguments one by one *)
  let compare_args :
      ?strict:bool ->
      ?level:Report.level ->
      S.pos ->
      argument list ->
      argument_repr list ->
      Report.t list ->
      Report.t list =
   fun ?(strict = false) ?(level = Report.Warn) pos expected actuals report ->
    let tl, report =
      List.fold_left actuals ~init:(expected, report)
        ~f:(fun (expected, report) param ->
          match expected with
          | (Variable _ as hd) :: _ ->
              let check = compare_parameter ~strict ~level hd param report in
              (expected, check)
          | hd :: tl ->
              let check = compare_parameter ~strict ~level hd param report in
              (tl, check)
          | [] ->
              let msg = Report.error param.pos "Unexpected argument" in
              ([], msg :: report))
    in
    match tl with
    | [] | Variable _ :: _ -> report
    | _ ->
        let msg = Report.error pos "Not enougth arguments given" in
        msg :: report
end

module Expression = struct
  type t = { result : Helper.t; pos : S.pos; empty : bool }
  type t' = t

  let v t = t

  let arg_of_repr : t -> Helper.argument_repr =
   fun { result; pos; empty } ->
    ignore empty;
    { pos; t = result }

  (** The variable has type string when starting with a '$' *)
  let ident : (S.pos, t S.repr) S.variable -> t S.repr =
   fun var report ->
    let empty = false in
    match var.name.[0] with
    | '$' -> ({ result = Variable String; pos = var.pos; empty }, report)
    | _ -> ({ result = Variable Integer; pos = var.pos; empty }, report)

  let integer : S.pos -> string -> t S.repr =
   fun pos value report ->
    let int_value = int_of_string_opt value in

    let empty, report =
      match int_value with
      | Some 0 -> (true, report)
      | Some _ -> (false, report)
      | None -> (false, Report.error pos "Invalid integer value" :: report)
    in

    ({ result = Raw Integer; pos; empty }, report)

  let literal : S.pos -> string -> t S.repr =
   fun pos value report ->
    let empty = String.equal String.empty value in
    let type_of =
      match int_of_string_opt value with
      | Some _ -> Helper.NumericString
      | None -> Helper.String
    in
    ({ result = Raw type_of; pos; empty }, report)

  let function_ : S.pos -> T.function_ -> t S.repr list -> t S.repr =
   fun pos function_ params _acc ->
    (* Accumulate the expressions and get the results, the report is given in
       the differents arguments, and we build a list with the type of the
       parameters. *)
    let types, report =
      List.fold_left params ~init:([], _acc) ~f:(fun (types, report) param ->
          let t, report = param report in
          let arg = arg_of_repr t in
          (arg :: types, report))
    in
    let types = List.rev types
    and default = { result = Variable NumericString; pos; empty = false } in

    match function_ with
    | Arrcomp | Arrpos | Arrsize | Countobj | Desc | Dyneval | Getobj | Instr
    | Isplay ->
        ({ default with result = Variable Integer }, report)
    | Desc' | Dyneval' | Getobj' ->
        ({ default with result = Variable String }, report)
    | Func | Func' -> ({ default with result = Variable NumericString }, report)
    | Iif | Iif' ->
        let d = Helper.DynType.t () in
        let expected = Helper.[ Fixed Bool; Dynamic d; Dynamic d ] in
        let report = Helper.compare_args pos expected types report in
        (* Extract the type for the expression *)
        let result = d (Raw Helper.Bool) in
        ({ result; pos; empty = false }, report)
    | Input | Input' ->
        (* Input should check the result if the variable is a num and raise a
           message in this case.*)
        let expected = Helper.[ Fixed String ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Variable NumericString; pos; empty = false }, report)
    | Isnum ->
        let expected = Helper.[ Fixed String ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Bool; pos; empty = false }, report)
    | Lcase | Lcase' | Ucase | Ucase' ->
        let expected = Helper.[ Fixed String ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw String; pos; empty = false }, report)
    | Len ->
        let expected = Helper.[ Fixed NumericString ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Integer; pos; empty = false }, report)
    | Loc ->
        let expected = Helper.[ Fixed String ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Variable Bool; pos; empty = false }, report)
    | Max | Max' | Min | Min' ->
        let d = Helper.DynType.t () in
        (* All the arguments must have the same type *)
        let expected = Helper.[ Variable (Dynamic d) ] in
        let report = Helper.compare_args pos expected types report in
        let result = d (Raw Helper.Bool) in
        ({ result; pos; empty = false }, report)
    | Mid | Mid' ->
        let expected = Helper.[ Fixed String; Variable (Fixed Integer) ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Variable String; pos; empty = false }, report)
    | Msecscount -> ({ result = Raw Integer; pos; empty = false }, report)
    | Rand ->
        let expected = Helper.[ Variable (Fixed Integer) ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Integer; pos; empty = false }, report)
    | Replace -> ({ result = Variable String; pos; empty = false }, report)
    | Replace' -> ({ result = Variable String; pos; empty = false }, report)
    | Rgb -> ({ result = Raw Integer; pos; empty = false }, report)
    | Qspver | Qspver' | Rnd ->
        (* No arg *)
        let report = Helper.compare_args pos [] types report in
        ({ result = Raw Integer; pos; empty = false }, report)
    | Selact -> ({ result = Variable String; pos; empty = false }, report)
    | Stattxt -> ({ result = Variable String; pos; empty = false }, report)
    | Stattxt' -> ({ result = Variable String; pos; empty = false }, report)
    | Str | Str' ->
        let expected = Helper.[ Variable (Fixed Integer) ] in
        let report = Helper.compare_args pos expected types report in
        ({ default with result = Raw String }, report)
    | Strcomp -> ({ result = Raw Bool; pos; empty = false }, report)
    | Strfind -> ({ result = Variable String; pos; empty = false }, report)
    | Strfind' -> ({ result = Variable String; pos; empty = false }, report)
    | Strpos -> ({ result = Raw Integer; pos; empty = false }, report)
    | Trim -> ({ result = Variable String; pos; empty = false }, report)
    | Trim' -> ({ result = Variable String; pos; empty = false }, report)
    | Val ->
        let expected = Helper.[ Fixed NumericString ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Integer; pos; empty = false }, report)

  (** Unary operator like [-123] or [+'Text']*)
  let uoperator : S.pos -> T.uoperator -> t S.repr -> t S.repr =
   fun pos operator t1 report ->
    let t, report = t1 report in
    match operator with
    | Add -> (t, report)
    | Neg | No ->
        let types = [ arg_of_repr t ] in
        let expected = Helper.[ Fixed Integer ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Integer; pos; empty = false }, report)

  let boperator : S.pos -> T.boperator -> t S.repr -> t S.repr -> t S.repr =
   fun pos operator t1 t2 report ->
    let t1, report = t1 report in
    let t2, report = t2 report in
    let types = [ arg_of_repr t1; arg_of_repr t2 ] in
    match operator with
    | T.Plus ->
        let d = Helper.DynType.t () in
        (* Remove the empty elements *)
        let types =
          List.filter_map [ t1; t2 ] ~f:(fun t ->
              (* TODO could be added in the logs *)
              match t.empty with true -> None | false -> Some (arg_of_repr t))
        in
        let expected = List.map types ~f:(fun _ -> Helper.Dynamic d) in

        let report = Helper.compare_args pos expected types report in
        let result = d Helper.(Variable Integer) in
        ({ result; pos; empty = false }, report)
    | T.Eq | T.Neq ->
        (* If the expression is '' or 0, we accept the comparaison as if
            instead of raising a warning *)
        if t1.empty || t2.empty then
          ({ result = Raw Bool; pos; empty = false }, report)
        else
          let d = Helper.(Dynamic (DynType.t ())) in
          let expected = [ d; d ] in
          let report =
            Helper.compare_args ~strict:true pos expected (List.rev types)
              report
          in
          ({ result = Raw Bool; pos; empty = false }, report)
    | Lt | Gte | Lte | Gt ->
        let d = Helper.(Dynamic (DynType.t ())) in
        let expected = [ d; d ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Bool; pos; empty = false }, report)
    | T.Mod | T.Minus | T.Product | T.Div ->
        (* Operation over number *)
        let expected = Helper.[ Fixed Integer; Fixed Integer ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Integer; pos; empty = false }, report)
    | T.And | T.Or ->
        (* Operation over booleans *)
        let expected = Helper.[ Fixed Bool; Fixed Bool ] in
        let report = Helper.compare_args pos expected types report in
        ({ result = Raw Bool; pos; empty = false }, report)
end

module Instruction = struct
  type t = unit
  type t' = unit

  let v = Fun.id

  type expression = Expression.t' S.repr

  (** Call for an instruction like [GT] or [*CLR] *)
  let call : S.pos -> T.keywords -> expression list -> t S.repr =
   fun _pos _ expressions report ->
    List.fold_left expressions ~init:((), report)
      ~f:(fun ((), report) expression ->
        let result, report = expression report in
        ignore result;
        ((), report))

  let location : S.pos -> string -> t S.repr = fun _pos _ report -> ((), report)

  (** Comment *)
  let comment : S.pos -> t S.repr = fun _pos report -> ((), report)

  (** Raw expression *)
  let expression : expression -> t S.repr =
   fun expression report -> ((), snd (expression report))

  (** Helper function used in the [if_] function. *)
  let fold_clause :
      t * Report.t list -> (expression, t) S.clause -> t * Report.t list =
   fun ((), report) (_pos, expr, instructions) ->
    let result, report = expr report in
    let report =
      Helper.compare Helper.Bool (Expression.arg_of_repr result) report
    in
    List.fold_left instructions ~init:((), report)
      ~f:(fun ((), report) instruction -> instruction report)

  let if_ :
      S.pos ->
      (expression, t) S.clause ->
      elifs:(expression, t) S.clause list ->
      else_:(S.pos * t S.repr list) option ->
      t S.repr =
   fun _pos clause ~elifs ~else_ report ->
    (* Traverse the whole block recursively *)
    let report = fold_clause ((), report) clause in
    let report = List.fold_left elifs ~f:fold_clause ~init:report in

    match else_ with
    | None -> report
    | Some (_, instructions) ->
        List.fold_left instructions ~init:report
          ~f:(fun ((), report) instruction -> instruction report)

  let act : S.pos -> label:expression -> t S.repr list -> t S.repr =
   fun _pos ~label instructions report ->
    let result, report = label report in
    let report =
      Helper.compare Helper.String (Expression.arg_of_repr result) report
    in
    List.fold_left instructions ~init:((), report)
      ~f:(fun ((), report) instruction -> instruction report)

  let assign :
      S.pos ->
      (S.pos, expression) S.variable ->
      T.assignation_operator ->
      expression ->
      t S.repr =
   fun pos variable _ expression report ->
    let right_expression, report = expression report in
    match right_expression.empty with
    | true -> ((), report)
    | false -> (
        let expr1, report = Expression.ident variable report in
        let op1 = Expression.arg_of_repr expr1 in
        let op2 = Expression.arg_of_repr right_expression in

        let d = Helper.DynType.t () in
        (* Every part of the assignation should be the same type *)
        let expected = Helper.[ Dynamic d; Dynamic d ] in

        match
          Helper.compare_args ~strict:false ~level:Report.Error pos expected
            [ op1; op2 ] []
        with
        | [] ->
            ( (),
              Helper.compare_args ~strict:true ~level:Report.Debug pos expected
                [ op1; op2 ] report )
        | reports -> ((), reports @ report))
end

module Location = struct
  type t = unit
  type instruction = Instruction.t

  let location : S.pos -> instruction S.repr list -> t S.repr =
   fun _pos instructions report ->
    let (), report =
      List.fold_left instructions ~init:((), report)
        ~f:(fun ((), report) instruction -> instruction report)
    in
    ((), report)
end
